<div {% if widget._center_content==true %}
  style="display: flex; gap: {{{widget._gap}}}px; align-items: {{{widget._vertical_alignment}}}; justify-content: center;"
  {% else %} style="display: flex; gap: {{{widget._gap}}}px; align-items: {{{widget._vertical_alignment}}};" {% endif%}>
  {% for w in widget._widgets %}
  {{{w}}}
  {% endfor %}
</div>