<div class="fflex">
    {{{ widget._first_button }}}
    {{{ widget._prev_button }}}
    <div class="ml10 mr10">
        {{state.{{{widget.widget_id}}}.currentGrid}} / {{state.{{{widget.widget_id}}}.totalGrids}}{{state.{{{widget.widget_id}}}.sliderTitle}}
    </div>
    {{{ widget._next_button }}}
    {{{ widget._last_button }}}
  <div style="flex: 1 1 auto;">{{{widget._slider}}}</div>
</div>
{{{widget._grid_gallery}}}
